from datetime import datetime
#from kervi.messaging.email import EmailHandler
from kervi.messaging.user_log import UserLogPlugin
from kervi.config import Configuration
from kervi.controllers import Controller
from kervi.actions import action
from kervi.plugin.plugin_manager import PluginManager
from kervi.plugin.messaging.message_plugin import MessagePlugin
from kervi.core.authentication import Authorization 

class MessageManager(Controller):
    def __init__(self):
        Controller.__init__(self, "message_manager", "Message handler")
        self.spine.register_command_handler("messageManagerSend", self.send_message)
        self._channels = {}
        self._authorization = Authorization()
        self._plugin_manager = None
        self._plugin_manager = PluginManager(Configuration, "messaging", [MessagePlugin])
        self._users = self._authorization.get_users()
        self.load()

    def load(self):

        self._plugin_manager.add_plugin(UserLogPlugin(Configuration, self))
        for plugin in self._plugin_manager.plugins:
            self._channels[plugin.message_type] = plugin

        self._config = Configuration.messaging
        self._levels = Configuration.log.levels

    def add_channel(self, channel_id, handler):
        self._channels[channel_id] = handler

    #@action
    def send_message(self, subject, **kwargs):
        message_channels = kwargs.pop("channels", self._config.default_channels)
        groups = kwargs.pop("user_groups", [])
        timestamp = (datetime.utcnow() - datetime(1970, 1, 1)).total_seconds()
        time = datetime.utcnow()
        kwargs = dict(kwargs, time=time, timestamp=timestamp)
        if not groups:
            users = self._users
        else:
            users = []
            for user in self._users:
                ingroup = any(i in groups for i in user.groups)
                if ingroup:
                    users += [user]
        #print("u", users, message_channels, self._channels.keys())
        if users:
            for message_channel in message_channels:
                
                if message_channel in self._channels.keys():
                    channel = self._channels[message_channel]
                    channel_users = []
                    if channel.address_based:
                        for user in users:
                            if user.addresses.get(message_channel, None):
                                channel_users.append(user)
                        
                        if channel_users:
                            channel.send_message(channel_users, subject, **kwargs)
                    else:
                        channel.send_message([], subject, **kwargs)





class KerviPlugin(object):
    def __init__(self, name, config, manager):
        self._name = name
        self._config = config
        self._manager = manager

    @property
    def manager(self):
        return self._manager

    @property
    def config(self):
        return self._config

    @property
    def name(self):
        return self._name



from kervi.plugin.messaging.message_plugin import MessagePlugin
from kervi.config import Configuration
import kervi.spine as spine

class UserLogPlugin(MessagePlugin):
    def __init__(self, config, manager):
        MessagePlugin.__init__(self, "user_log", config, manager)
        #self._config = Configuration.messaging.channels.user_log
        self.spine = spine.Spine()

    @property
    def message_type(self):
        return "user_log"

    @property
    def address_based(self):
        return False

    def create_address(self, user):
        pass

    def send_message(self, addresses, subject, **kwargs):
        """"
        Adds a message to the user log.
        """
        persist = kwargs.get("persist", False)
        body = kwargs.get("body", None)
        data = kwargs.get("data", None)
        area = kwargs.get("area", None)
        source_id = kwargs.get("source_id", None)
        source_name = kwargs.get("source_name", None)
        log_type = kwargs.get("log_type", "information")
        level = kwargs.get("level", 3)
        timestamp = kwargs.get("timestamp", 0)
        self.spine.trigger_event("userLogMessage", None, {
            "source_id": source_id,
            "source_name": source_name,
            "area": area,
            "data": data,
            "level": level,
            "topic": subject,
            "body": body,
            "timestamp": timestamp,
            "type": log_type,
            "persist": persist
        })






import smtplib
from email.headerregistry import Address
from email.mime.multipart import MIMEMultipart
from email.mime.text import MIMEText
from kervi.plugin.messaging.message_plugin import MessagePlugin
from kervi.config import Configuration
from kervi.core.utility.superformatter import SuperFormatter
import datetime
from email.utils import formatdate

class EmailPlugin(MessagePlugin):
    def __init__(self, config, manager):
        MessagePlugin.__init__(self, "email", config, manager)

    @property
    def message_type(self):
        return "email"

    @property
    def address_based(self):
        return True
    
    def send_message(self, recipients, topic, **kwargs):
        body = kwargs.get("body", None)
        body_html = kwargs.get("body_html", None)
        source_name = kwargs.get("source_name", None)
        message_type = kwargs.get("message_type", None)

        recipients_addresses = []
        for recipient in recipients:
            email = recipient.addresses.get("email", None)
            if email:
                recipients_addresses.append(email)

        subject = ""
        if message_type:
            subject += "[" + message_type + "] "

        if source_name:
            subject += source_name + ": "

        subject += topic


        for recipient in recipients:
            email = recipient.addresses.get("email", None)
            msg = MIMEMultipart('alternative')
            msg['Subject'] = subject
            msg['Date'] = formatdate(localtime=True)
            msg['From'] = self._config.smtp.sender_name + "<" + self._config.smtp.sender_address + ">"
            msg['To'] = recipient.name + "<" + email + ">"

            if body:
                sf = SuperFormatter()
                body_text = sf.format(body, user_name=recipient.name)
                part1 = MIMEText(body_text, 'plain')
                msg.attach(part1)
            if body_html:
                part2 = MIMEText(body_html, 'html')
                msg.attach(part2)

            
            with smtplib.SMTP(self._config.smtp.server, self._config.smtp.port) as smtp:
                if self._config.smtp.tls:
                    smtp.starttls()

                if self._config.smtp.user and self._config.smtp.password:
                    smtp.login(self._config.smtp.user, self._config.smtp.password)

                smtp.send_message(msg)
       
def init_plugin(config, manager):
    return EmailPlugin(config, manager)


import inspect
from kervi.config.configuration import _KerviConfig
class PluginManager:
    def __init__(self, config, section, plugin_classes=None):
        self._config = config
        self._plugins = []
        self._plugin_classes = plugin_classes
        for plugin in config.plugins[section].keys:
            plugin_config = config
            if isinstance(config.plugins[section][plugin], _KerviConfig):
                enabled = config.plugins[section][plugin]["enabled"]
                plugin_config = config.plugins[section][plugin]
            else:
                enabled = config.plugins[section][plugin]

            if enabled:
                module = __import__(plugin, fromlist=[''])
                plugin = module.init_plugin(plugin_config, self)
                if self.add_plugin(plugin):
                    print("loaded plugin:", module.__name__)
                else:
                    print("Invalid plugin class:", type(plugin))

    @property
    def config(self):
        return self._config
    
    def add_plugin(self, plugin):
        is_valid = True
        if self._plugin_classes:
            #plugin_bases = inspect.getmro(type(plugin))
            is_valid = False
            for plugin_class in self._plugin_classes:
                if isinstance(plugin, plugin_class):
                    is_valid = True
                    break
        if is_valid:
            self._plugins.append(plugin)
            return True
        else:
            return False

    @property
    def plugins(self):
        return self._plugins



class _UserAddresses:
    def __init__(self, addresses):
        self._addresses = addresses

    def get(self, address_type, default_value):
        
        if self._addresses and address_type in self._addresses.keys:
            return self._addresses[address_type]
        return default_value

class KerviUser(object):
    def __init__(self, user_name, name, groups, addresses):
        self.user_name = user_name
        if name:
            self.name = name
        else:
            self.name = user_name
        self.groups = groups
        self._addresses = _UserAddresses(addresses)
        
    @property
    def addresses(self):
        return self._addresses

from kervi.plugin import KerviPlugin
class AuthenticationPlugin(KerviPlugin):
    def __init__(self, name, configuration, manager):
        KerviPlugin.__init__(self, name, configuration, manager)

    def authorize(self, user_name, password):
        raise NotImplementedError

    def get_users(self):
        raise NotImplementedError


#MIT License
#Copyright (c) 2017 Tim Wentzlau

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Module that holds classes for creating / bootstraping a Kervi application or a Kervi module.
"""
import os
import time
import threading
try:
    import thread
except ImportError:
    import _thread as thread

import sys
if sys.version_info[0] < 3:
    print("Kervi requires Python 3")
    exit()

#from kervi.application.kervi_module import KerviModule
#import kervi.core.utility.process as process
#import kervi.spine as spine
#from kervi.utility.process_spine import _ProcessSpine
#import kervi.kervi_logging as logging
import kervi.utility.nethelper as nethelper
#import kervi.utility.encryption as encryption
from kervi.controllers import Controller
from kervi.actions import action
from kervi.application.default_config import get_default_config

import signal

_app_running = True

def handler_stop_signals(signum, frame):
    global _app_running
    _app_running = False
    print("signal:", signum)

def _pretty_print(d, indent=0):
    if isinstance(d, dict):
        for key in d.keys():
            value = d[key]
            
            if isinstance(value, dict) or isinstance(value, list):
                print( '  ' * indent + str(key))
                _pretty_print(value, indent+1)
            else:
                print('  ' * (indent+1) + key + ":" +  str(value))
    elif isinstance(d, list):
        for item in d:
            if isinstance(item, dict) or isinstance(item, list):
                _pretty_print(item, indent+1)
            else:
                print('  ' * (indent+1) + str(item))
    else:
        pass    
    

class _AppActions(Controller):
    def __init__(self, app):
        super().__init__("app", "App controller")

        self._app = app

    @action
    def shutdown(self):
        print("shutdown app device")
        self._app._shutdown_action()

    @action
    def reboot(self):
        print("reboot app device")
        self._app._reboot_action()

    @action
    def restart(self):
        print("restart app")
        self._app._restart_action()

    @action
    def stop(self):
        print("stop app")
        self._app._stop_action()

class Application(object):
    """ Kervi application class that starts a kervi application and loads all needed modules.
        This class should be used by it self like:

    """
    def __init__(self, user_config = None):
        """ Settings is a dictionary with the following content
        """
        
        
        print("Starting kervi application")
        import inspect
        import getopt
        
        config_files = []

        opts, args = getopt.getopt(sys.argv[1:], "c", ["config_file=", "as-service", "install-service", "uninstall-service", "start-service", "stop-service", "restart-service", "status-service", "detect-devices"])
        for opt, arg in opts:
            if opt in ("-c", "--config_file"):
                if os.path.isfile(arg):
                    config_files += [arg]
                else:
                    print("Specified config file not found:", arg)
        script_path = os.path.abspath(inspect.stack()[1][1])
        script_name = os.path.basename(script_path)
        script_name, script_ext = os.path.splitext(script_name)
        
        config_files += [script_name +".config.json"]
        config_files += ["kervi.config.json"]

        selected_config_file = None
        for config_file in config_files:
            if os.path.isfile(config_file):
                selected_config_file = config_file
                break
        #if not selected_config_file:
        #    print("no config file found , revert to defaults")

        from kervi.config import load
        self.config = load(
            config_file=selected_config_file,
            config_user=user_config,
            config_base=get_default_config()
        )

        service_commands = []
        detect_devices = None
        self._as_service = False
        for opt, arg in opts:
            if opt in ("--as-service"):
                self._as_service = True

            if opt in ("--detect-devices"):
                detect_devices = True

            if opt in ("--install-service"):
                service_commands += ["install"]

            if opt in ("--uninstall-service"):
                service_commands += ["uninstall"]

            if opt in ("--start-service"):
                service_commands += ["start"]

            if opt in ("--stop-service"):
                service_commands = ["stop"]
            
            if opt in ("--restart-service"):
                service_commands = ["restart"]
            if opt in ("--service-status"):
                service_commands = ["status"]

        if service_commands:
            import kervi.hal as hal
            hal_driver = hal._load(self.config.platform.driver)
            if hal_driver:
                hal.service_commands(
                    service_commands,
                    self.config.application.name,
                    self.config.application.id,
                    script_path
                )
            exit()

        if detect_devices:
            import kervi.hal as hal
            hal_driver = hal._load(self.config.platform.driver)
            if hal_driver:
                devices = hal.detect_devices()
                print("devices:")
                _pretty_print(devices)
            exit()

        
        #if settings:
        #    self.settings = app_helpers._deep_update(self.settings, settings)
        #self._validateSettings()
        self.started = False
        import kervi.spine as spine
        from kervi.zmq_spine import _ZMQSpine
        self.spine = _ZMQSpine()
        self.config.network.ipc_root_port = nethelper.get_free_port([self.config.network.ipc_root_port])
        self.spine._init_spine("kervi-main", self.config.network.ipc_root_port, None, self.config.network.ipc_root_address)
        spine.set_spine(self.spine)
        #process._start_root_spine(self.config, True, _ZMQSpine)
        #spine._init_spine("application-" + self.settings["info"]["id"])
        self.spine = spine.Spine()
        self.spine.register_query_handler("GetApplicationInfo", self._get_application_info)
        self.spine.register_query_handler("getProcessInfo", self.get_process_info)
        self.spine.register_event_handler("modulePing", self._module_ping)
        self.spine.register_event_handler("processReady", self._process_ready, scope="app-" + self.config.application.id)
        self._module_processes = []
        self._process_info = []
        self._process_info_lock = threading.Lock()

        self._kervi_modules = []
        self._kervi_modules_lock = threading.Lock()

        #from kervi.utility.storage import init_db
        #init_db(script_name)
        from kervi.storage.storage_manager import StorageManager
        self._authentication = StorageManager()
        
        from kervi.utility.authorization_manager import AuthorizationManager
        self._authentication = AuthorizationManager()
        
        from kervi.messaging.message_manager import MessageManager
        self._message_handler = MessageManager()

        self._app_actions = _AppActions(self)

        import kervi.hal as hal
        hal_driver = hal._load(self.config.platform.driver)
        if hal_driver:
            print("platform driver:", hal_driver)


    @property
    def actions(self):
        return self._app_actions

    def _validateSettings(self):
        pass

    def _get_application_info(self):
        unit_system = self.config.unit_system
        display = {}
        display["unit_system"] = self.config.display.unit_systems.systems[unit_system]
        display["texts"] = self.config.texts.ui

        return {
            "name": self.config.application.name,
            "id": self.config.application.id,
            "display": display,
            "unit_system": unit_system
        }

    def get_process_info(self):
        return {"id": "application"}

    def _module_ping(self, module_id):
        with self._kervi_modules_lock:
            for module in self._kervi_modules:
                if module.module_id == module_id:
                    
                    if not module.is_alive:
                        print("module started", module_id)
                        self.spine.trigger_event(
                            "moduleStarted",
                            module_id
                        )
                    module.ping()
                    break
            else:
                print("module started", module_id)
                self._kervi_modules += [KerviModule(module_id)]
                self.spine.trigger_event(
                    "moduleStarted",
                    module_id
                )

    def _check_kervi_modules(self):
        with self._kervi_modules_lock:
            for module in self._kervi_modules:
                if not module.is_alive and module.connected:
                    print("module stopped", module.module_id)
                    module.connected = False
                    self.spine.trigger_event(
                        "moduleStopped",
                        module.module_id
                    )

    def _process_ready(self, scope, process_id):
        self._process_info_lock.acquire()
        try:
            for process in self._process_info:
                if process["id"] == process_id:
                    process["ready"] = True
        finally:
            self._process_info_lock.release()

    def _is_ready(self):
        result = True
        self._process_info_lock.acquire()
        try:
            for process in self._process_info:
                if not process["ready"]:
                    result = False
        finally:
            self._process_info_lock.release()

        return result

    def _start(self):
        self.started = True

        try:
            import dashboards
        except ImportError:
            pass
        
        import kervi.core.utility.process as process
        import kervi.utility.application_helpers as app_helpers


        if self._as_service:
            signal.signal(signal.SIGINT, handler_stop_signals)
            signal.signal(signal.SIGTERM, handler_stop_signals)

        self.spine.send_command("startThreads", local_only=True)
        time.sleep(.5)
        module_port = self.config.network.ipc_root_port

        self._process_info_lock.acquire()
        self._process_info = [{"id":"IPC", "ready":False}]
        self._process_info_lock.release()

        module_port += 1
        self._module_processes += [
            process._start_process(
                "app-" + self.config.application.id,
                "IPC",
                self.config,
                nethelper.get_free_port([module_port]),
                app_helpers._KerviSocketIPC
            )
        ]

        for module in self.config.modules:
            self._process_info_lock.acquire()
            self._process_info += [{"id":module, "ready":False}]
            self._process_info_lock.release()

            module_port += 1
            self._module_processes += [
                process._start_process(
                    "app-" + self.config.application.id,
                    module,
                    self.config,
                    nethelper.get_free_port([module_port]),
                    app_helpers._KerviModuleLoader
                )
            ]

        # if self.config.routing.kervi_io.enabled:
        #     module_port += 1
        #     self._module_processes += [
        #         process._start_process(
        #             "app-" + self.config.application.id,
        #             "kervi_io",
        #             self.config,
        #             nethelper.get_free_port([module_port]),
        #             app_helpers._KerviIORouterProcess
        #         )
        #     ]

        while not self._is_ready():
            time.sleep(1)

        #self._module_processes += app_helpers.load_plugins(self.config, module_port)

        from kervi.dashboards import Dashboard
        Dashboard._add_default()

        import platform
        if platform.system() != "Windows":
            print("\033[92mYour Kervi application is ready at http://" + self.config.network.ip + ":" + str(self.config.network.http_port) + "\033[0m")
        else:
            print("Your Kervi application is ready at http://" + self.config.network.ip + ":" + str(self.config.network.http_port))
        print("Press ctrl + c to stop your application")
        import kervi.ui.webserver as webserver
        webserver.start(
            self.config.network.ip,
            self.config.network.http_port,
            self.config.network.ws_port
        )
        self.spine.trigger_event(
            "appReady",
            self.config.application.id
        )
        self.spine.send_command("startWebSocket")
        self.spine.send_command("kervi_action_app_main")

    def _input_thread(self, list):
        try:
            raw_input()
            list.append(None)
        except KeyboardInterrupt:
            pass
        except:
            pass

    def run(self):
        """Run the application loop. The application will continue until termination with ctrl-c"""

        if not self.started:
            self._start()
        try:
            char_list = []
            thread.start_new_thread(self._input_thread, (char_list,))
            while not char_list and _app_running:
                self._check_kervi_modules()
                time.sleep(1)

        except KeyboardInterrupt:
            pass

        self.stop()

    def _xrun(self):
        """Used in tests"""

        if not self.started:
            self._start()

    def stop(self, force_exit=True, restart=False):
        self.spine.send_command("kervi_action_app_exit")
        import kervi.ui.webserver as webserver
        
        webserver.stop()
        print("stopping processes")
        import kervi.core.utility.process as process
        process._stop_processes("app-" + self.config.application.id)
        self.spine.trigger_event("processTerminating", None, local_only=True)
        time.sleep(1)
        self.spine.stop()
        #for thread in threading.enumerate():
        #    print("running thread",thread.name)
        print("application stopped")
        #exit()
        if force_exit:
            import os
            os._exit(0)

    def _stop_action(self):
        self.stop()

    def _restart_action(self):
        self.stop(restart=True)

    def _reboot_action(self):
        import kervi.hal as hal
        hal.device_reboot()

    def _shutdown_action(self):
        import kervi.hal as hal
        hal.device_shutdown()



import socket 
import sys
import threading
import sys
import time
import json

class KerviAppDiscovery(threading.Thread):
	def __init__(self, ip, port, discovery_port, app_id, challenge, app_name, web_address):
		threading.Thread.__init__(self, None, None, "Kervi App discovery")
		self.deamon = True
		self._terminate = False
		self._sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
		server_address = ("0.0.0.0", discovery_port)
		self._sock.bind(server_address)
		self._response = "I am a kervi app with id:" + app_id
		self._hello = "Are you a kervi app with id: " + app_id
		self._detect = "Are you a kervi app with challenge: " + challenge
		self._detect_response = "I am a kervi app with challenge: " + challenge
		
		self._port = port
		self._discovery_port = discovery_port
		self._app_id = app_id
		self._app_name = app_name
		self._web_address = web_address
		self._ip = ip

	def terminate(self):
		self._terminate = True
		self._sock.close()

	def run(self):
		print("Listining for discovery requests on port:", self._discovery_port)
		while not self._terminate:
			try:
				data, address = self._sock.recvfrom(4096)
				data = str(data.decode('UTF-8'))
				#print('Received ' + str(len(data)) + ' bytes from ' + str(address) )
				#print('Data:' + data)
				if data == self._hello:
					response = {
						"challenge": self._response,
						"ip": self._ip,
						"port": self._port
					}
					response_data = json.dumps(response)
					self._sock.sendto(response_data.encode(), address)
				
				if data == self._detect:
					response = {
						"challenge": self._detect_response,
						"port": self._port,
						"id": self._app_id,
						"name": self._app_name,
						"web": self._web_address
					}
					response_data = json.dumps(response)
					self._sock.sendto(response_data.encode(), address)
				
					#print('Sent confirmation back', response)
			except OSError:
				self._terminate = True
		#print("discovery terminated")


def find_kervi_app(app_id, discovery_port=9434):
	result = (None, None)
	# Create a UDP socket
	sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
	sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
	sock.setsockopt(socket.SOL_SOCKET, socket.SO_BROADCAST, 1)
	sock.settimeout(2)

	use_local = True
	message = "Are you a kervi app with id: " + app_id
	message_reply = "I am a kervi app with id:" + app_id
	try:
		while True:
			# Send data
			#print('sending: ' + message)
			if use_local:
				server_address = ('127.255.255.255', discovery_port)
			else:
				server_address = ('255.255.255.255', discovery_port)
			use_local = not use_local
			sock.sendto(message.encode(), server_address)

			# Receive response
			#print('waiting to receive')
			try:
				data, server = sock.recvfrom(4096)
				response = json.loads(data.decode("UTF-8"))
				#print("response", server[0], response)
				if response["challenge"] == message_reply:
					result = (response["ip"], response["port"])
					#print('Received confirmation')
					#print('Server ip: ', str(server[0]), response["port"])
					break
				else:
					print('Challenge verification failed.')
			
				#print('Trying again...')
			except KeyboardInterrupt:
				break
			except TimeoutError:
				pass
			except Exception as ex:
				#print("e", ex)
				pass
		
	except KeyboardInterrupt:
		pass
	finally:	
		sock.close()

	return result


from kervi.plugin.authentication.authentication_plugin import AuthenticationPlugin, KerviUser

class PlainAuthenticationPlugin(AuthenticationPlugin):
    def __init__(self, configuration, manager):
        AuthenticationPlugin.__init__(self, "plain", configuration, manager)
        self._users = manager.config.plain_users
    def allow_anonymous(self):
        if "anonymous" in self._users.keys:
            return self._users.anonymous.enabled
        return False
    
    def _get_user(self, user_name, config_user):
        
        return KerviUser(
            user_name,
            config_user.get("name", user_name),
            config_user.get("groups", None),
            config_user.get("addresses", {})
        )

    def authorize(self, user_name, password):
        users = self._users
        user = users.get(user_name, None)
        if user:
            if user_name == "anonymous":
                if self.allow_anonymous():
                    return self._get_user(user_name, user)
            else:
                user_password = user.get("password", None)
                enabled = user.get("enabled", False)
                if user_password and enabled and user_password == password:
                    return self._get_user(user_name, user)
                elif not user_password and not password:
                    return self._get_user(user_name, user)
        return None

    def get_users(self):
        result=[]
        users = self._users
        for user_name in users.keys:
            user = users[user_name]
            result.append(self._get_user(user_name, user))
        return result

def init_plugin(config, manager):
    return PlainAuthenticationPlugin(config, manager)